@@ -1,5 +1,5 @@
 __docformat__ = "restructuredtext"
-__version__ = "0.5.4"
+__version__ = "0.6.2"
 
 __doc__ = """
 `data_algebra`<https://github.com/WinVector/data_algebra> is a piped data wrangling system
@@ -13,13 +13,13 @@ are subject to change.
 
 Mature, production ready `R`<https://www.r-project.org> versions of the system are available as 
 the `rquery`<https://github.com/WinVector/rquery> and `rqdatatable`<https://github.com/WinVector/rqdatatable> packages.
-
-Recommended packages include: Pandas, PyYAML (supplies yaml), sqlparse, and black. 
 """
 
 import data_algebra.pandas_model
-import pandas
+
+# set up what pandas supplier we are using
+import pandas as pd
 
 default_data_model = data_algebra.pandas_model.PandasModel(
-    pd=pandas, presentation_model_name="pandas"
+    pd=pd, presentation_model_name="pd"
 )
