@@ -1,19 +1,21 @@
 from abc import ABC
 from typing import Set, Dict, List, Union
 import numbers
-import collections
 import re
+import collections
 
 import data_algebra
+import data_algebra.expr_parse
 import data_algebra.flow_text
 import data_algebra.data_model
 import data_algebra.db_model
 import data_algebra.pandas_model
 import data_algebra.expr_rep
-import data_algebra.env
+import data_algebra.user_fn
 from data_algebra.data_ops_types import *
 import data_algebra.data_ops_utils
 import data_algebra.near_sql
+import data_algebra.OrderedSet
 
 _have_black = False
 try:
@@ -34,37 +36,13 @@ except ImportError:
     pass
 
 
-# wrap a single argument function as a user callable function in pipeline
-# used for custom aggregators
-def user_fn(fn, args=None, *, name=None):
-    if isinstance(fn, str):
-        if name is None:
-            name = fn
-        fn = eval(fn)
-    if not callable(fn):
-        raise TypeError("expected fn to be callable")
-    if args is None:
-        args = []
-    if isinstance(args, str):
-        fn_args = [data_algebra.expr_rep.ColumnReference(view=None, column_name=args)]
-    else:
-        for v in args:
-            if not isinstance(v, str):
-                raise TypeError("Expect all vars names to be strings")
-        fn_args = [
-            data_algebra.expr_rep.ColumnReference(view=None, column_name=v)
-            for v in args
-        ]
-    return data_algebra.expr_rep.FnCall(fn, fn_args=fn_args, name=name)
-
-
 class ViewRepresentation(OperatorPlatform, ABC):
     """Structure to represent the columns of a query or a table.
        Abstract base class."""
 
     column_names: List[str]
     column_set: Set[str]
-    column_map: data_algebra.env.SimpleNamespaceDict
+    column_map: collections.OrderedDict
     sources: List[
         "ViewRepresentation"
     ]  # https://www.python.org/dev/peps/pep-0484/#forward-references
@@ -78,14 +56,16 @@ class ViewRepresentation(OperatorPlatform, ABC):
                 raise ValueError("non-string column name(s)")
         if len(self.column_names) < 1:
             raise ValueError("no column names")
-        self.column_set = set(self.column_names)
+        self.column_set = data_algebra.OrderedSet.OrderedSet()
+        for c in self.column_names:
+            self.column_set.add(c)
         if not len(self.column_names) == len(self.column_set):
             raise ValueError("duplicate column name(s)")
         column_dict = {
             ci: data_algebra.expr_rep.ColumnReference(self, ci)
             for ci in self.column_names
         }
-        self.column_map = data_algebra.env.SimpleNamespaceDict(**column_dict)
+        self.column_map = collections.OrderedDict(**column_dict)
         if sources is None:
             sources = []
         for si in sources:
@@ -97,16 +77,6 @@ class ViewRepresentation(OperatorPlatform, ABC):
     def merged_rep_id(self):
         return "node+ " + str(id(self))
 
-    # adaptors
-
-    def get_column_symbols(self):
-        """Return a representation of this step as columns we can perform algebraic operations over.
-        These objects capture the operations as an expression tree."""
-        column_defs = self.column_map.__dict__
-        nd = column_defs.copy()
-        ns = data_algebra.env.SimpleNamespaceDict(**nd)
-        return ns
-
     # characterization
 
     def get_tables(self):
@@ -192,20 +162,6 @@ class ViewRepresentation(OperatorPlatform, ABC):
                 have.update(f)
         return res
 
-    # collect as simple structures for YAML I/O and other generic tasks
-
-    def collect_representation_implementation(self, *, pipeline=None, dialect="Python"):
-        raise NotImplementedError("base method called")
-
-    def collect_representation(self, *, pipeline=None, dialect="Python"):
-        """Collect a representation of the operator DAG as simple serializable objects.
-                   These objects can be saved/loaded in YAML format and also can rebuild the
-                   pipeline via data_algebra.yaml.to_pipeline()."""
-        self.columns_used()  # for table consistency check/raise
-        return self.collect_representation_implementation(
-            pipeline=pipeline, dialect=dialect
-        )
-
     # printing
 
     def to_python_implementation(self, *, indent=0, strict=True, print_sources=True):
@@ -282,9 +238,11 @@ class ViewRepresentation(OperatorPlatform, ABC):
     ):
         if sqlparse_options is None:
             sqlparse_options = {"reindent": True, "keyword_case": "upper"}
+        if isinstance(db_model, data_algebra.db_model.DBHandle):
+            db_model = db_model.db_model
         if not isinstance(db_model, data_algebra.db_model.DBModel):
             raise TypeError(
-                "Expected db_model to be derived from data_algebra.db_model.DBModel"
+                "Expected db_model to be derived from data_algebra.db_model.DBModel or data_algebra.db_model.DBHandle"
             )
         self.columns_used()  # for table consistency check/raise
         temp_id_source = [0]
@@ -485,7 +443,7 @@ class ViewRepresentation(OperatorPlatform, ABC):
     ):
         if (ops is None) or (len(ops) < 1):
             return self
-        parsed_ops = data_algebra.expr_rep.parse_assignments_in_context(
+        parsed_ops = data_algebra.expr_parse.parse_assignments_in_context(
             ops, self, parse_env=parse_env
         )
         return self.extend_parsed(
@@ -512,12 +470,19 @@ class ViewRepresentation(OperatorPlatform, ABC):
             group_by = []
         if ((ops is None) or (len(ops) < 1)) and (len(group_by) < 1):
             raise ValueError("must have ops or group_by")
-        parsed_ops = data_algebra.expr_rep.parse_assignments_in_context(
+        parsed_ops = data_algebra.expr_parse.parse_assignments_in_context(
             ops, self, parse_env=parse_env
         )
         return self.project_parsed(parsed_ops=parsed_ops, group_by=group_by)
 
-    def natural_join(self, b, *, by=None, jointype="INNER"):
+    def natural_join(self, b, *, by, jointype):
+        """
+
+        :param b: right table
+        :param by: list of keys to join by
+        :param jointype: one of 'INNER', 'LEFT', 'RIGHT', 'OUTER', 'FULL', 'CROSS' (case insensitive)
+        :return: ops describing join
+        """
         if not isinstance(b, ViewRepresentation):
             raise TypeError(
                 "expected b to be a data_algebra.dat_ops.ViewRepresentation"
@@ -553,9 +518,25 @@ class ViewRepresentation(OperatorPlatform, ABC):
             return self
         if self.is_trivial_when_intermediate():
             return self.sources[0].select_rows(expr, parse_env=parse_env)
-        ops = data_algebra.expr_rep.parse_assignments_in_context(
+        ops = data_algebra.expr_parse.parse_assignments_in_context(
             {"expr": expr}, self, parse_env=parse_env
         )
+        data_model = data_algebra.default_data_model
+        forbid = data_model.custom_function_map.keys()
+
+        def r_walk_expr(op):
+            if not isinstance(op, data_algebra.expr_rep.Expression):
+                return
+            if op.op in forbid:
+                raise ValueError("special functions such as "
+                                 + op.op
+                                 + " not allowed in select_rows.\n"
+                                 + " Work around: please use an extend to materialize this expression.")
+            for oi in op.args:
+                r_walk_expr(oi)
+
+        for op in ops.values():
+            r_walk_expr(op)
         return self.select_rows_parsed(parsed_expr=ops)
 
     def drop_columns(self, column_deletions):
@@ -592,12 +573,12 @@ class ViewRepresentation(OperatorPlatform, ABC):
             return self.sources[0].order_rows(columns, reverse=reverse, limit=limit)
         return OrderRowsNode(source=self, columns=columns, reverse=reverse, limit=limit)
 
-    def convert_records(self, record_map):
+    def convert_records(self, record_map, *, temp_namer=None):
         if record_map is None:
             return self
         if self.is_trivial_when_intermediate():
-            return self.sources[0].convert_records(record_map)
-        return ConvertRecordsNode(source=self, record_map=record_map)
+            return self.sources[0].convert_records(record_map, temp_namer=temp_namer)
+        return ConvertRecordsNode(source=self, record_map=record_map, temp_namer=temp_namer)
 
 
 # Could also have general query as starting node, but don't see a lot of point to
@@ -607,15 +588,9 @@ class ViewRepresentation(OperatorPlatform, ABC):
 class TableDescription(ViewRepresentation):
     """Describe columns, and qualifiers, of a table.
 
-       If outer namespace is set user values are visible and
-       _-side effects can be written back.
-
        Example:
            from data_algebra.data_ops import *
-           import data_algebra.env
-           with data_algebra.env.Env(globals()) as env:
-               d = TableDescription('d', ['x', 'y'])
-           print(_) # should be a SimpleNamespaceDict, not d/ViewRepresentation
+           d = TableDescription('d', ['x', 'y'])
            print(d)
     """
 
@@ -630,6 +605,7 @@ class TableDescription(ViewRepresentation):
         column_names,
         *,
         qualifiers=None,
+        sql_meta=None,
         column_types=None,
         head=None,
         limit_was=None
@@ -646,6 +622,7 @@ class TableDescription(ViewRepresentation):
                 raise ValueError("head.columns != column_names")
         self.head = head
         self.limit_was = limit_was
+        self.sql_meta = sql_meta
         self.table_name = table_name
         if isinstance(column_names, str):
             column_names = [column_names]
@@ -658,17 +635,9 @@ class TableDescription(ViewRepresentation):
         if not isinstance(qualifiers, dict):
             raise TypeError("qualifiers must be a dictionary")
         self.qualifiers = qualifiers.copy()
-        key = ""
-        if len(self.qualifiers) > 0:
-            keys = [k for k in self.qualifiers.keys()]
-            keys.sort()
-            key = "{"
-            for k in keys:
-                key = key + "(" + k + ", " + str(self.qualifiers[k]) + ")"
-            key = key + "}."
+        self.key = ""
         if self.table_name is not None:
-            key = key + self.table_name
-        self.key = key
+            self.key = self.table_name
 
     def same_table(self, other):
         if not isinstance(other, data_algebra.data_ops.TableDescription):
@@ -718,18 +687,6 @@ class TableDescription(ViewRepresentation):
             return False
         return True
 
-    def collect_representation_implementation(self, *, pipeline=None, dialect="Python"):
-        if pipeline is None:
-            pipeline = []
-        od = collections.OrderedDict()
-        od["op"] = "TableDescription"
-        od["table_name"] = self.table_name
-        od["qualifiers"] = self.qualifiers.copy()
-        od["column_names"] = self.column_names
-        od["key"] = self.key
-        pipeline.insert(0, od)
-        return pipeline
-
     def to_python_implementation(self, *, indent=0, strict=True, print_sources=True):
         spacer = "\n " + " " * indent
         column_limit = 20
@@ -804,25 +761,33 @@ class TableDescription(ViewRepresentation):
         return self.key.__hash__()
 
 
-def describe_table(d, table_name="data_frame", *, qualifiers=None, column_types=None):
+def describe_table(d, table_name="data_frame",
+                   *,
+                   qualifiers=None,
+                   sql_meta=None,
+                   column_types=None,
+                   row_limit=7):
     # Expect a pandas.DataFrame style object
+    assert d is not None
+    assert not isinstance(d, OperatorPlatform)
+    assert not isinstance(d, ViewRepresentation)
     column_names = [c for c in d.columns]
     if column_types is None:
         if d.shape[0] > 0:
             column_types = {k: type(d.loc[0, k]) for k in column_names}
-    limit_was = 6
-    if d.shape[0] > limit_was:
-        head = d.iloc[range(limit_was), :]
+    if d.shape[0] > row_limit:
+        head = d.iloc[range(row_limit), :].copy()
     else:
         head = d.copy()
-    head = head.reset_index(drop=True, inplace=False)
+    head.reset_index(drop=True, inplace=True)
     return TableDescription(
         table_name,
         column_names,
         column_types=column_types,
         qualifiers=qualifiers,
+        sql_meta=sql_meta,
         head=head,
-        limit_was=limit_was,
+        limit_was=row_limit,
     )
 
 
@@ -985,20 +950,6 @@ class ExtendNode(ViewRepresentation):
             o.get_column_names(columns_we_take)
         return [columns_we_take]
 
-    def collect_representation_implementation(self, *, pipeline=None, dialect="Python"):
-        if pipeline is None:
-            pipeline = []
-        od = collections.OrderedDict()
-        od["op"] = "Extend"
-        od["ops"] = {ci: vi.to_source(dialect=dialect) for (ci, vi) in self.ops.items()}
-        od["partition_by"] = self.partition_by
-        od["order_by"] = self.order_by
-        od["reverse"] = self.reverse
-        pipeline.insert(0, od)
-        return self.sources[0].collect_representation_implementation(
-            pipeline=pipeline, dialect=dialect
-        )
-
     def to_python_implementation(self, *, indent=0, strict=True, print_sources=True):
         spacer = "\n   " + " " * indent
         s = ""
@@ -1087,7 +1038,7 @@ class ProjectNode(ViewRepresentation):
                             + str(opk)
                         )
             else:
-                if not isinstance(opk, data_algebra.expr_rep.FnCall):
+                if not isinstance(opk, data_algebra.user_fn.FnTerm):
                     raise ValueError(
                         "non-aggregated expression in project: "
                         + str(k)
@@ -1133,18 +1084,6 @@ class ProjectNode(ViewRepresentation):
             o.get_column_names(columns_we_take)
         return [columns_we_take]
 
-    def collect_representation_implementation(self, *, pipeline=None, dialect="Python"):
-        if pipeline is None:
-            pipeline = []
-        od = collections.OrderedDict()
-        od["op"] = "Project"
-        od["ops"] = {ci: vi.to_source(dialect=dialect) for (ci, vi) in self.ops.items()}
-        od["group_by"] = self.group_by
-        pipeline.insert(0, od)
-        return self.sources[0].collect_representation_implementation(
-            pipeline=pipeline, dialect=dialect
-        )
-
     def to_python_implementation(self, *, indent=0, strict=True, print_sources=True):
         spacer = "\n   " + " " * indent
         s = ""
@@ -1225,17 +1164,6 @@ class SelectRowsNode(ViewRepresentation):
         columns_we_take = columns_we_take.union(self.decision_columns)
         return [columns_we_take]
 
-    def collect_representation_implementation(self, *, pipeline=None, dialect="Python"):
-        if pipeline is None:
-            pipeline = []
-        od = collections.OrderedDict()
-        od["op"] = "SelectRows"
-        od["expr"] = self.expr.to_source(dialect=dialect)
-        pipeline.insert(0, od)
-        return self.sources[0].collect_representation_implementation(
-            pipeline=pipeline, dialect=dialect
-        )
-
     def to_python_implementation(self, *, indent=0, strict=True, print_sources=True):
         s = ""
         if print_sources:
@@ -1268,9 +1196,11 @@ class SelectColumnsNode(ViewRepresentation):
             columns = [columns]
         column_selection = [c for c in columns]
         self.column_selection = column_selection
+        if len(column_selection) < 1:
+            raise ValueError("can not drop all columns")
         unknown = set(column_selection) - set(source.column_names)
         if len(unknown) > 0:
-            raise ValueError("selecting unknown columns " + str(unknown))
+            raise KeyError("selecting unknown columns " + str(unknown))
         if isinstance(source, SelectColumnsNode):
             source = source.sources[0]
         ViewRepresentation.__init__(
@@ -1305,17 +1235,6 @@ class SelectColumnsNode(ViewRepresentation):
             return [cols]
         return [cols.intersection(using)]
 
-    def collect_representation_implementation(self, *, pipeline=None, dialect="Python"):
-        if pipeline is None:
-            pipeline = []
-        od = collections.OrderedDict()
-        od["op"] = "SelectColumns"
-        od["columns"] = self.column_selection
-        pipeline.insert(0, od)
-        return self.sources[0].collect_representation_implementation(
-            pipeline=pipeline, dialect=dialect
-        )
-
     def to_python_implementation(self, *, indent=0, strict=True, print_sources=True):
         s = ""
         if print_sources:
@@ -1348,9 +1267,14 @@ class DropColumnsNode(ViewRepresentation):
             column_deletions = [column_deletions]
         column_deletions = [c for c in column_deletions]
         self.column_deletions = column_deletions
+        unknown = set(column_deletions) - set(source.column_names)
+        if len(unknown) > 0:
+            raise KeyError("dropping unknown columns " + str(unknown))
         remaining_columns = [
             c for c in source.column_names if c not in column_deletions
         ]
+        if len(remaining_columns) < 1:
+            raise ValueError("can not drop all columns")
         ViewRepresentation.__init__(
             self,
             column_names=remaining_columns,
@@ -1382,17 +1306,6 @@ class DropColumnsNode(ViewRepresentation):
             using = set(self.sources[0].column_names)
         return [set([c for c in using if c not in self.column_deletions])]
 
-    def collect_representation_implementation(self, *, pipeline=None, dialect="Python"):
-        if pipeline is None:
-            pipeline = []
-        od = collections.OrderedDict()
-        od["op"] = "DropColumns"
-        od["column_deletions"] = self.column_deletions
-        pipeline.insert(0, od)
-        return self.sources[0].collect_representation_implementation(
-            pipeline=pipeline, dialect=dialect
-        )
-
     def to_python_implementation(self, *, indent=0, strict=True, print_sources=True):
         s = ""
         if print_sources:
@@ -1471,19 +1384,6 @@ class OrderRowsNode(ViewRepresentation):
         cols = cols.intersection(using).union(self.order_columns)
         return [cols]
 
-    def collect_representation_implementation(self, *, pipeline=None, dialect="Python"):
-        if pipeline is None:
-            pipeline = []
-        od = collections.OrderedDict()
-        od["op"] = "Order"
-        od["order_columns"] = self.order_columns
-        od["reverse"] = self.reverse
-        od["limit"] = self.limit
-        pipeline.insert(0, od)
-        return self.sources[0].collect_representation_implementation(
-            pipeline=pipeline, dialect=dialect
-        )
-
     def to_python_implementation(self, *, indent=0, strict=True, print_sources=True):
         s = ""
         if print_sources:
@@ -1584,17 +1484,6 @@ class RenameColumnsNode(ViewRepresentation):
         ]
         return [set(cols)]
 
-    def collect_representation_implementation(self, *, pipeline=None, dialect="Python"):
-        if pipeline is None:
-            pipeline = []
-        od = collections.OrderedDict()
-        od["op"] = "Rename"
-        od["column_remapping"] = self.column_remapping
-        pipeline.insert(0, od)
-        return self.sources[0].collect_representation_implementation(
-            pipeline=pipeline, dialect=dialect
-        )
-
     def to_python_implementation(self, *, indent=0, strict=True, print_sources=True):
         s = ""
         if print_sources:
@@ -1621,10 +1510,12 @@ class NaturalJoinNode(ViewRepresentation):
     by: List[str]
     jointype: str
 
-    def __init__(self, a, b, *, by=None, jointype="INNER"):
+    def __init__(self, a, b, *, by, jointype):
         # check set of tables is consistent in both sub-dags
         a_tables = a.get_tables()
         b_tables = b.get_tables()
+        if by is None:
+            raise ValueError("Must specify by in natural joins ([] for empty conditions)")
         common_keys = set(a_tables.keys()).intersection(b_tables.keys())
         for k in common_keys:
             if not a_tables[k].same_table(b_tables[k]):
@@ -1656,6 +1547,8 @@ class NaturalJoinNode(ViewRepresentation):
         )
         self.by = by
         self.jointype = data_algebra.expr_rep.standardize_join_type(jointype)
+        if (self.jointype == 'CROSS') and (len(self.by) != 0):
+            raise ValueError("CROSS joins must have an empty 'by' list")
         self.get_tables()  # causes a throw if left and right table descriptions are inconsistent
 
     def apply_to(self, a, *, target_table_key=None):
@@ -1681,19 +1574,6 @@ class NaturalJoinNode(ViewRepresentation):
         using = using.union(self.by)
         return [self.sources[i].column_set.intersection(using) for i in range(2)]
 
-    def collect_representation_implementation(self, *, pipeline=None, dialect="Python"):
-        if pipeline is None:
-            pipeline = []
-        od = collections.OrderedDict()
-        od["op"] = "NaturalJoin"
-        od["by"] = self.by
-        od["jointype"] = self.jointype
-        od["b"] = self.sources[1].collect_representation_implementation(dialect=dialect)
-        pipeline.insert(0, od)
-        return self.sources[0].collect_representation_implementation(
-            pipeline=pipeline, dialect=dialect
-        )
-
     def to_python_implementation(self, *, indent=0, strict=True, print_sources=True):
         s = "_0."
         if print_sources:
@@ -1788,18 +1668,6 @@ class ConcatRowsNode(ViewRepresentation):
             return [self.sources[i].column_set.copy() for i in range(2)]
         return [self.sources[i].column_set.intersection(using) for i in range(2)]
 
-    def collect_representation_implementation(self, *, pipeline=None, dialect="Python"):
-        if pipeline is None:
-            pipeline = []
-        od = collections.OrderedDict()
-        od["op"] = "ConcatRows"
-        od["id_column"] = self.id_column
-        od["b"] = self.sources[1].collect_representation_implementation(dialect=dialect)
-        pipeline.insert(0, od)
-        return self.sources[0].collect_representation_implementation(
-            pipeline=pipeline, dialect=dialect
-        )
-
     def to_python_implementation(self, *, indent=0, strict=True, print_sources=True):
         s = "_0."
         if print_sources:
@@ -1844,9 +1712,10 @@ class ConcatRowsNode(ViewRepresentation):
 
 
 class ConvertRecordsNode(ViewRepresentation):
-    def __init__(self, source, record_map):
+    def __init__(self, *, source, record_map, temp_namer=None):
         sources = [source]
         self.record_map = record_map
+        self.temp_namer = temp_namer
         unknown = set(self.record_map.columns_needed) - set(source.column_names)
         if len(unknown) > 0:
             raise ValueError("missing required columns: " + str(unknown))
@@ -1857,13 +1726,15 @@ class ConvertRecordsNode(ViewRepresentation):
             node_name="ConvertRecordsNode",
         )
 
-    def blocks_out_table(self, *, temp_id_source):
-        view_name = "cdata_temp_record_" + str(temp_id_source[0])
+    def control_out_table(self, *, temp_id_source):
+        if self.temp_namer is None:
+            view_name = "cdata_temp_record_" + str(temp_id_source[0])
+        else:
+            view_name = self.temp_namer(temp_id_source[0])
         temp_id_source[0] = temp_id_source[0] + 1
         res = TableDescription(
-            view_name,
-            [c for c in self.record_map.blocks_out.record_keys]
-            + [c for c in self.record_map.blocks_out.control_table.columns],
+            table_name=view_name,
+            column_names=self.record_map.blocks_out.control_table.columns
         )
         return res
 
@@ -1883,17 +1754,6 @@ class ConvertRecordsNode(ViewRepresentation):
     def columns_used_from_sources(self, using=None):
         return [self.record_map.columns_needed]
 
-    def collect_representation_implementation(self, *, pipeline=None, dialect="Python"):
-        if pipeline is None:
-            pipeline = []
-        od = collections.OrderedDict()
-        od["op"] = "ConvertRecords"
-        od["record_map"] = self.record_map.to_simple_obj()
-        pipeline.insert(0, od)
-        return self.sources[0].collect_representation_implementation(
-            pipeline=pipeline, dialect=dialect
-        )
-
     def to_python_implementation(self, *, indent=0, strict=True, print_sources=True):
         s = ""
         if print_sources:
@@ -1917,19 +1777,19 @@ class ConvertRecordsNode(ViewRepresentation):
         )
         # claims to use all columns
         query = sub_query.to_sql(
-            columns=self.columns_used_from_sources()[0], db_model=db_model
+            columns=self.columns_used_from_sources()[0], db_model=db_model, force_sql=True
         )
-        blocks_out_table = None
+        control_out_table = None
         if self.record_map.blocks_in is not None:
             query = db_model.blocks_to_row_recs_query(
                 query, record_spec=self.record_map.blocks_in
             )
         if self.record_map.blocks_out is not None:
-            blocks_out_table = self.blocks_out_table(temp_id_source=temp_id_source)
+            control_out_table = self.control_out_table(temp_id_source=temp_id_source)
             query = db_model.row_recs_to_blocks_query(
                 query,
                 record_spec=self.record_map.blocks_out,
-                record_view=blocks_out_table,
+                control_view=control_out_table,
             )
         view_name = "convert_records_in_" + str(temp_id_source[0])
         temp_id_source[0] = temp_id_source[0] + 1
@@ -1937,12 +1797,12 @@ class ConvertRecordsNode(ViewRepresentation):
         temp_id_source[0] = temp_id_source[0] + 1
         terms = {k: None for k in self.record_map.columns_produced}
         temp_tables = sub_query.temp_tables.copy()
-        if blocks_out_table is not None:
-            if blocks_out_table.key in temp_tables.keys():
+        if control_out_table is not None:
+            if control_out_table.key in temp_tables.keys():
                 raise ValueError(
-                    "key collision in temp_tables construction: " + blocks_out_table.key
+                    "key collision in temp_tables construction: " + control_out_table.key
                 )
-            temp_tables[blocks_out_table.key] = self.record_map.blocks_out.control_table
+            temp_tables[control_out_table.key] = self.record_map.blocks_out.control_table
         near_sql = data_algebra.near_sql.NearSQLq(
             quoted_query_name=db_model.quote_identifier(view_name),
             prev_quoted_query_name=db_model.quote_identifier(prev_view_name),
