@@ -1,11 +1,14 @@
 import math
 import re
 import io
+from types import SimpleNamespace
+from collections import OrderedDict
 
 import data_algebra
 
 import data_algebra.near_sql
 import data_algebra.expr_rep
+import data_algebra.user_fn
 import data_algebra.util
 import data_algebra.data_ops_types
 import data_algebra.eval_model
@@ -64,14 +67,77 @@ def _db_if_else_expr(dbmodel, expression):
     )
 
 
+def _db_maximum_expr(dbmodel, expression):
+    x_expr = dbmodel.expr_to_sql(expression.args[0], want_inline_parens=True)
+    y_expr = dbmodel.expr_to_sql(expression.args[1], want_inline_parens=True)
+    return (
+        "CASE"
+        + " WHEN "
+        + "(" + x_expr + ") >= (" + y_expr + ")"
+        + " THEN "
+        + x_expr
+        + " WHEN NOT "
+        + "(" + x_expr + ") >= (" + y_expr + ")"
+        + " THEN "
+        + y_expr
+        + " ELSE NULL END"
+    )
+
+
+def _db_minimum_expr(dbmodel, expression):
+    x_expr = dbmodel.expr_to_sql(expression.args[0], want_inline_parens=True)
+    y_expr = dbmodel.expr_to_sql(expression.args[1], want_inline_parens=True)
+    return (
+        "CASE"
+        + " WHEN "
+        + "(" + x_expr + ") <= (" + y_expr + ")"
+        + " THEN "
+        + x_expr
+        + " WHEN NOT "
+        + "(" + x_expr + ") <= (" + y_expr + ")"
+        + " THEN "
+        + y_expr
+        + " ELSE NULL END"
+    )
+
+
+def _db_is_in_expr(dbmodel, expression):
+    is_in_expr = dbmodel.expr_to_sql(expression.args[0], want_inline_parens=True)
+    x_expr = dbmodel.expr_to_sql(expression.args[1], want_inline_parens=True)
+    return (
+        "("
+        + is_in_expr
+        + " IN ("
+        + ', '.join([repr(v) for v in x_expr])
+        + "))"
+    )
+
+
+def _db_count_expr(dbmodel, expression):
+    return 'SUM(1)'
+
+
 db_expr_formatters = {
     "is_null": _db_is_null_expr,
     "is_bad": _db_is_bad_expr,
     "if_else": _db_if_else_expr,
+    "is_in": _db_is_in_expr,
+    "maximum": _db_maximum_expr,
+    "minimum": _db_minimum_expr,
+    'count': _db_count_expr,
 }
 
 
-db_default_op_replacements = {"==": "=", "cumsum": "sum"}
+db_default_op_replacements = {
+    "==": "=",
+    "cumsum": "sum",
+    "&": "AND",
+    "&&": "AND",
+    "|": "OR",
+    "||": "OR",
+    "!": "NOT",
+    "~": "NOT",
+    }
 
 
 class DBModel:
@@ -80,6 +146,12 @@ class DBModel:
 
     identifier_quote: str
     string_quote: str
+    on_start: str
+    on_end: str
+    on_joiner: str
+    drop_text: str
+    string_type: str
+    join_name_map: dict
 
     def __init__(
         self,
@@ -88,7 +160,13 @@ class DBModel:
         string_quote="'",
         sql_formatters=None,
         op_replacements=None,
-        local_data_model=None
+        local_data_model=None,
+        on_start='',
+        on_end='',
+        on_joiner=', ',
+        drop_text='DROP TABLE',
+        string_type='VARCHAR',
+        join_name_map=None,
     ):
         if local_data_model is None:
             local_data_model = data_algebra.default_data_model
@@ -104,6 +182,17 @@ class DBModel:
         if op_replacements is None:
             op_replacements = db_default_op_replacements
         self.op_replacements = op_replacements
+        self.on_start = on_start
+        self.on_end = on_end
+        self.on_joiner = on_joiner
+        self.drop_text = drop_text
+        self.string_type = string_type
+        if join_name_map is None:
+            join_name_map = {}
+        self.join_name_map = join_name_map.copy()
+
+    def db_handle(self, conn):
+        return DBHandle(db_model=self, conn=conn)
 
     def prepare_connection(self, conn):
         pass
@@ -129,12 +218,12 @@ class DBModel:
             + (
                 "double precision"
                 if self.local_data_model.can_convert_col_to_numeric(d[d.columns[i]])
-                else "VARCHAR"
+                else self.string_type
             )
             for i in range(d.shape[1])
         ]
-        q_table_name = self.build_qualified_table_name(
-            table_name, qualifiers=qualifiers
+        q_table_name = self.quote_table_name(
+            table_name
         )
         cur = conn.cursor()
         # check for table
@@ -148,7 +237,7 @@ class DBModel:
             if not allow_overwrite:
                 raise ValueError("table " + q_table_name + " already exists")
             else:
-                cur.execute("DROP TABLE " + q_table_name)
+                cur.execute(self.drop_text + ' ' + q_table_name)
                 conn.commit()
         create_stmt = "CREATE TABLE " + q_table_name + " ( " + ", ".join(cr) + " )"
         cur.execute(create_stmt)
@@ -161,9 +250,13 @@ class DBModel:
     def execute(self, conn, q):
         """
 
-        :param conn: database connection
+        :param conn: database connectionex
         :param q: sql query
         """
+        if isinstance(q, data_algebra.data_ops.ViewRepresentation):
+            q = q.to_sql(db_model=self)
+        else:
+            q = str(q)
         cur = conn.cursor()
         cur.execute(q)
 
@@ -174,6 +267,10 @@ class DBModel:
         :param q: sql query
         :return: query results as table
         """
+        if isinstance(q, data_algebra.data_ops.ViewRepresentation):
+            q = q.to_sql(db_model=self)
+        else:
+            q = str(q)
         cur = conn.cursor()
         cur.execute(q)
         r = cur.fetchall()
@@ -186,8 +283,8 @@ class DBModel:
     def read_table(self, conn, table_name, *, qualifiers=None, limit=None):
         if not isinstance(table_name, str):
             raise TypeError("Expect table_name to be a str")
-        q_table_name = self.build_qualified_table_name(
-            table_name, qualifiers=qualifiers
+        q_table_name = self.quote_table_name(
+            table_name
         )
         sql = "SELECT * FROM " + q_table_name
         if limit is not None:
@@ -207,17 +304,11 @@ class DBModel:
         if not isinstance(identifier, str):
             raise TypeError("expected identifier to be a str")
         if self.identifier_quote in identifier:
-            raise TypeError('did not expect " in identifier')
+            # TODO: escape quote
+            # TODO: unify with derived PostgreSQL and BigQuery
+            raise TypeError('did not expect ' + self.identifier_quote + ' in identifier')
         return self.identifier_quote + identifier + self.identifier_quote
 
-    def build_qualified_table_name(self, table_name, *, qualifiers=None):
-        qt = self.quote_identifier(table_name)
-        if qualifiers is None:
-            qualifiers = {}
-        if len(qualifiers) > 0:
-            raise ValueError("This data model does not expect table qualifiers")
-        return qt
-
     def quote_table_name(self, table_description):
         if isinstance(table_description, str):
             return self.quote_identifier(table_description)
@@ -225,9 +316,7 @@ class DBModel:
             raise TypeError(
                 "Expected table_description to be a data_algebra.data_ops.TableDescription)"
             )
-        return self.build_qualified_table_name(
-            table_description.table_name, qualifiers=table_description.qualifiers
-        )
+        return self.quote_identifier(table_description.table_name)
 
     def quote_string(self, string):
         if not isinstance(string, str):
@@ -245,6 +334,10 @@ class DBModel:
     def value_to_sql(self, v):
         if v is None:
             return "NULL"
+        if isinstance(v, data_algebra.expr_rep.ListTerm):
+            return [self.value_to_sql(vi) for vi in v.value]
+        if isinstance(v, data_algebra.expr_rep.Value):
+            return v.value
         if isinstance(v, str):
             return self.quote_string(v)
         if isinstance(v, bool):
@@ -256,6 +349,10 @@ class DBModel:
             if math.isnan(v):
                 return "NULL"
             return str(v)
+        if isinstance(v, int):
+            return str(v)
+        if isinstance(v, list) or isinstance(v, tuple):
+            return [vi for vi in v]
         return str(v)
 
     def expr_to_sql(self, expression, *, want_inline_parens=False):
@@ -283,26 +380,8 @@ class DBModel:
                     # SQL window functions don't like parens
                     return subs[0] + " " + op.upper() + " " + subs[1]
             return op.upper() + "(" + ", ".join(subs) + ")"
-        if isinstance(expression, data_algebra.expr_rep.FnCall):
-            op = expression.name
-            if op in self.op_replacements.keys():
-                op = self.op_replacements[op]
-            if op in self.sql_formatters.keys():
-                return self.sql_formatters[op](self, expression)
-            subs = [
-                self.expr_to_sql(ai, want_inline_parens=True) for ai in expression.args
-            ]
-            return op.upper() + "(" + ", ".join(subs) + ")"
-        if isinstance(expression, data_algebra.expr_rep.FnValue):
-            op = expression.name
-            if op in self.op_replacements.keys():
-                op = self.op_replacements[op]
-            if op in self.sql_formatters.keys():
-                return self.sql_formatters[op](self, expression)
-            subs = [
-                self.expr_to_sql(ai, want_inline_parens=True) for ai in expression.args
-            ]
-            return op.upper() + "(" + ", ".join(subs) + ")"
+        if isinstance(expression, data_algebra.user_fn.FnTerm):
+            return expression.to_sql(db_model=self)
         raise TypeError("unexpected type: " + str(type(expression)))
 
     def table_def_to_sql(self, table_def, *, using=None, temp_id_source=None):
@@ -314,19 +393,30 @@ class DBModel:
             temp_id_source = [0]
         if using is None:
             using = table_def.column_set
-        if len(using) < 1:
-            raise ValueError("must select at least one column")
         missing = using - table_def.column_set
         if len(missing) > 0:
             raise KeyError("referred to unknown columns: " + str(missing))
         cols_using = [c for c in table_def.column_names if c in using]
         view_name = "table_reference_" + str(temp_id_source[0])
         temp_id_source[0] = temp_id_source[0] + 1
-        near_sql = data_algebra.near_sql.NearSQLTable(
+        subsql = data_algebra.near_sql.NearSQLTable(
             terms={k: self.quote_identifier(k) for k in cols_using},
             quoted_query_name=self.quote_identifier(view_name),
             quoted_table_name=self.quote_table_name(table_def),
         )
+        near_sql = subsql
+        if (len(using) > 0) and (not (set([k for k in using]) == set([k for k in table_def.column_names]))):
+            # need a non-trivial select here
+            terms = OrderedDict()
+            for k in using:
+                terms[k] = k
+            near_sql = data_algebra.near_sql.NearSQLUnaryStep(
+                previous_step_summary=subsql.summary(),
+                terms=terms,  # TODO: implement!
+                quoted_query_name=self.quote_identifier(view_name),
+                sub_sql=subsql.to_sql(columns=using, db_model=self),
+                temp_tables=subsql.temp_tables.copy(),
+            )
         return near_sql
 
     def extend_to_sql(self, extend_node, *, using=None, temp_id_source=None):
@@ -341,7 +431,10 @@ class DBModel:
         using = using.union(
             extend_node.partition_by, extend_node.order_by, extend_node.reverse
         )
-        subops = {k: op for (k, op) in extend_node.ops.items() if k in using}
+        subops = OrderedDict()
+        for (k, op) in extend_node.ops.items():
+            if k in using:
+                subops[k] = op
         if len(subops) <= 0:
             # know using was not None is this case as len(extend_node.ops)>0 and all keys are in extend_node.column_set
             return extend_node.sources[0].to_sql_implementation(
@@ -375,10 +468,12 @@ class DBModel:
                 ]
                 window_term = window_term + "ORDER BY " + ", ".join(rt) + " "
             window_term = window_term + " ) "
-        terms = {ci: self.expr_to_sql(oi) + window_term for (ci, oi) in subops.items()}
-        origcols = {k: None for k in using if k not in subops.keys()}
-        if len(origcols) > 0:
-            terms.update(origcols)
+        terms = OrderedDict()
+        origcols = [k for k in using if k not in subops.keys()]
+        for k in origcols:
+            terms[k] = None
+        for (ci, oi) in subops.items():
+            terms[ci] = self.expr_to_sql(oi) + window_term
         view_name = "extend_" + str(temp_id_source[0])
         temp_id_source[0] = temp_id_source[0] + 1
         near_sql = data_algebra.near_sql.NearSQLUnaryStep(
@@ -401,8 +496,6 @@ class DBModel:
             using = project_node.column_set
         subops = {k: op for (k, op) in project_node.ops.items() if k in using}
         subusing = project_node.columns_used_from_sources(using=using)[0]
-        if (len(project_node.group_by) + len(subusing)) < 1:
-            raise ValueError("must use at least one column")
         terms = {ci: self.expr_to_sql(oi) for (ci, oi) in subops.items()}
         terms.update({g: None for g in project_node.group_by})
         subsql = project_node.sources[0].to_sql_implementation(
@@ -469,7 +562,7 @@ class DBModel:
         subsql = select_columns_node.sources[0].to_sql_implementation(
             db_model=self, using=set(subusing), temp_id_source=temp_id_source
         )
-        # see if we can order columns
+        # order/limit columns
         subsql.terms = {
             k: subsql.terms[k]
             for k in select_columns_node.column_selection
@@ -492,6 +585,10 @@ class DBModel:
         subsql = drop_columns_node.sources[0].to_sql_implementation(
             db_model=self, using=subusing, temp_id_source=temp_id_source
         )
+        # /limit columns
+        subsql.terms = {
+            k: v for k, v in subsql.terms.items() if k not in drop_columns_node.column_deletions
+        }
         return subsql
 
     def order_to_sql(self, order_node, *, using=None, temp_id_source=None):
@@ -578,13 +675,12 @@ class DBModel:
         if using is None:
             using = join_node.column_set
         by_set = set(join_node.by)
-        using = using.union(by_set)
         if len(using) < 1:
             raise ValueError("must select at least one column")
         missing = using - join_node.column_set
         if len(missing) > 0:
             raise KeyError("referred to unknown columns: " + str(missing))
-        subusing = join_node.columns_used_from_sources(using=using)
+        subusing = join_node.columns_used_from_sources(using=using.union(by_set))
         using_left = subusing[0]
         using_right = subusing[1]
         sql_left = join_node.sources[0].to_sql_implementation(
@@ -608,7 +704,7 @@ class DBModel:
             + "."
             + self.quote_identifier(ci)
             + ")"
-            for ci in common
+            for ci in common if ci in using
         }
         terms.update({ci: None for ci in using_left - common})
         terms.update({ci: None for ci in using_right - common})
@@ -616,7 +712,8 @@ class DBModel:
         if len(join_node.by) > 0:
             on_terms = (
                 " ON "
-                + ", ".join(
+                + self.on_start
+                + self.on_joiner.join(
                     [
                         sub_view_name_left
                         + "."
@@ -628,6 +725,7 @@ class DBModel:
                         for c in join_node.by
                     ]
                 )
+                + self.on_end
                 + " "
             )
         confused_temps = set(sql_left.temp_tables.keys()).intersection(
@@ -637,6 +735,11 @@ class DBModel:
             raise ValueError("name collisions on temp_tables: " + str(confused_temps))
         temp_tables = sql_left.temp_tables.copy()
         temp_tables.update(sql_right.temp_tables)
+        jointype = join_node.jointype
+        try:
+            jointype = self.join_name_map[jointype]
+        except KeyError:
+            pass
         near_sql = data_algebra.near_sql.NearSQLBinaryStep(
             terms=terms,
             quoted_query_name=self.quote_identifier(view_name),
@@ -717,7 +820,7 @@ class DBModel:
         return near_sql
 
     def row_recs_to_blocks_query(
-        self, source_sql, record_spec, record_view, *, using=None, temp_id_source=None
+        self, source_sql, record_spec, control_view, *, using=None, temp_id_source=None
     ):
         if temp_id_source is None:
             temp_id_source = [0]
@@ -725,9 +828,9 @@ class DBModel:
         #     raise TypeError(
         #         "record_spec should be a data_algebra.cdata.RecordSpecification"
         #     )
-        if not isinstance(record_view, data_algebra.data_ops_types.OperatorPlatform):
+        if not isinstance(control_view, data_algebra.data_ops_types.OperatorPlatform):
             raise TypeError(
-                "record_view should be a data_algebra.data_ops_types.OperatorPlatform"
+                "control_view should be a data_algebra.data_ops_types.OperatorPlatform"
             )
         control_value_cols = [
             c
@@ -766,7 +869,7 @@ class DBModel:
                     col_sql = (
                         "  WHEN CAST(b."
                         + self.quote_identifier(result_col)
-                        + " AS VARCHAR) = "
+                        + " AS " + self.string_type + ") = "
                         + self.quote_string(str(source_col))
                         + " THEN a."
                         + self.quote_identifier(source_col)
@@ -783,7 +886,7 @@ class DBModel:
             + source_sql
             + " ) a\n"
             + "CROSS JOIN (\n  "
-            + record_view.to_sql_implementation(
+            + control_view.to_sql_implementation(
                 self, using=using, temp_id_source=temp_id_source
             ).to_sql(db_model=self, columns=using, force_sql=True)
             + " ) b\n"
@@ -823,7 +926,7 @@ class DBModel:
                         clauses.append(
                             " ( CAST("
                             + self.quote_identifier(cc)
-                            + " AS VARCHAR) = "
+                            + " AS " + self.string_type + ") = "
                             + self.quote_string(str(record_spec.control_table[cc][i]))
                             + " ) "
                         )
@@ -852,9 +955,15 @@ class DBModel:
         )
         return sql
 
+    def __str__(self):
+        return str(type(self).__name__)
+
+    def __repr__(self):
+        return self.__str__()
+
 
 class DBHandle(data_algebra.eval_model.EvalModel):
-    def __init__(self, db_model, conn):
+    def __init__(self, *, db_model, conn, fns = None):
         if not isinstance(db_model, DBModel):
             raise TypeError(
                 "expected db_model to be of class data_algebra.db_model.DBHandle"
@@ -862,20 +971,35 @@ class DBHandle(data_algebra.eval_model.EvalModel):
         data_algebra.eval_model.EvalModel.__init__(self)
         self.db_model = db_model
         self.conn = conn
+        if fns is None:
+            fns = {}
+        self.fns = SimpleNamespace(**fns)
 
-    def build_rep(self, table_name, *, row_limit=7):
-        head = self.db_model.read_query(
-            conn=self.conn,
+    def __enter__(self):
+        return self
+
+    def __exit__(self, type, value, traceback):
+        self.close()
+
+    def read_query(self, q):
+        if isinstance(q, data_algebra.data_ops.ViewRepresentation):
+            q = q.to_sql(self.db_model)
+        else:
+            q = str(q)
+        return self.db_model.read_query(conn=self.conn, q=q)
+
+    def describe_table(self, table_name, *, qualifiers=None, row_limit=7):
+        head = self.read_query(
             q="SELECT * FROM "
-            + self.db_model.quote_table_name(table_name)
-            + " LIMIT "
-            + str(row_limit),
+                + self.db_model.quote_table_name(table_name)
+                + " LIMIT "
+                + str(row_limit),
         )
-        return data_algebra.data_ops.TableDescription(
-            column_names=head.columns,
+        return data_algebra.data_ops.describe_table(
+            head,
             table_name=table_name,
-            head=head,
-            limit_was=row_limit,
+            qualifiers=qualifiers,
+            row_limit=row_limit
         )
 
     def to_pandas(self, handle, *, data_map=None):
@@ -909,7 +1033,30 @@ class DBHandle(data_algebra.eval_model.EvalModel):
         self.db_model.insert_table(
             conn=self.conn, d=d, table_name=table_name, allow_overwrite=allow_overwrite
         )
-        return self.build_rep(table_name)
+        return self.describe_table(table_name)
+
+    def execute(self, q):
+        if isinstance(q, data_algebra.data_ops.ViewRepresentation):
+            q = q.to_sql(self.db_model)
+        else:
+            q = str(q)
+        self.db_model.execute(conn=self.conn, q=q)
+
+    def to_sql(self, ops,
+               *,
+               pretty=False,
+               encoding=None,
+               sqlparse_options=None,
+               temp_tables=None):
+        return ops.to_sql(self.db_model,
+                          pretty=pretty,
+                          encoding=encoding,
+                          sqlparse_options=sqlparse_options,
+                          temp_tables=temp_tables)
+
+    def drop_table(self, table_name):
+        q_table_name = self.db_model.quote_table_name(table_name)
+        self.execute(f'DROP TABLE IF EXISTS {q_table_name}')
 
     def eval(self, ops, *, data_map=None, result_name=None, eval_env=None, narrow=True):
         query = ops.to_sql(self.db_model)
@@ -939,7 +1086,7 @@ class DBHandle(data_algebra.eval_model.EvalModel):
         if result_name in tables_needed:
             raise ValueError("Can not write over an input table")
         q_table_name = self.db_model.quote_table_name(result_name)
-        drop_query = "DROP TABLE " + q_table_name
+        drop_query = self.db_model.drop_text + ' ' + q_table_name
         create_query = "CREATE TABLE " + q_table_name + " AS " + query
         cur = self.conn.cursor()
         # noinspection PyBroadException
@@ -948,14 +1095,15 @@ class DBHandle(data_algebra.eval_model.EvalModel):
         except Exception:
             pass
         cur.execute(create_query)
-        res = self.build_rep(result_name)
+        res = self.describe_table(result_name)
         if data_map is not None:
             data_map[result_name] = res
         return result_name
 
     def __str__(self):
         return (
-            "data_algebra.db_model.DBHandle("
+            str(type(self).__name__) +
+            "("
             + "db_model="
             + str(self.db_model)
             + ", conn="
@@ -965,3 +1113,8 @@ class DBHandle(data_algebra.eval_model.EvalModel):
 
     def __repr__(self):
         return self.__str__()
+
+    def close(self):
+        if self.conn is not None:
+            self.conn.close()
+            self.conn = None
