@@ -1,9 +1,7 @@
 from abc import ABC
 from typing import Union
-import collections
 
 import data_algebra.util
-import data_algebra.env
 import data_algebra.custom_functions
 
 # for some ideas in capturing expressions in Python see:
@@ -83,7 +81,7 @@ class PreTerm(ABC):
     # emitters
 
     def to_python(self, *, want_inline_parens=False):
-        raise NotImplementedError("base class called")
+        raise NotImplementedError("base class method called")  # https://docs.python.org/3/library/exceptions.html
 
     def to_pandas(self, *, want_inline_parens=False):
         return self.to_python(want_inline_parens=want_inline_parens)
@@ -107,7 +105,10 @@ class PreTerm(ABC):
 
 class Term(PreTerm, ABC):
     """Inherit from this class to capture expressions.
-    Abstract class, should be extended for use."""
+    Abstract class, should be extended for use.
+
+    Used by the parse by lark path via method lookup.
+    """
 
     def __init__(self):
         PreTerm.__init__(self)
@@ -146,7 +147,7 @@ class Term(PreTerm, ABC):
             y = _enc_value(y)
         return Expression(op, (self, x, y), inline=inline, method=method)
 
-    # try to get at == and other comparision operators
+    # try to get at == and other comparison operators
 
     def __eq__(self, other):
         return self.__op_expr__("==", other)
@@ -211,13 +212,12 @@ class Term(PreTerm, ABC):
         return self.__rop_expr__("**", other)
 
     def __neg__(self):
-        # return self.__uop_expr__("neg")
         return self.__uop_expr__("-", inline=True)
 
     def __pos__(self):
         return self  # Treat as a no-op
 
-    # # TODO: need to work out how to send to Pandas
+    # # TODO: need to work out how to send to Pandas and SQL
     # def __lshift__(self, other):
     #     return self.__op_expr__("<<", other)
     #
@@ -229,25 +229,26 @@ class Term(PreTerm, ABC):
     #
     # def __rrshift__(self, other):
     #     return self.__rop_expr__(">>", other)
-    #
-    # def __and__(self, other):
-    #     return self.__op_expr__("&", other)
-    #
-    # def __rand__(self, other):
-    #     return self.__rop_expr__("&", other)
-    #
-    # def __xor__(self, other):
-    #     return self.__op_expr__("^", other)
-    #
-    # def __rxor__(self, other):
-    #     return self.__rop_expr__("^", other)
-    #
-    # def __or__(self, other):
-    #     return self.__op_expr__("|", other)
-    #
-    # def __ror__(self, other):
-    #     return self.__rop_expr__("|", other)
 
+    def __and__(self, other):
+        return self.__op_expr__("&", other)
+
+    def __rand__(self, other):
+        return self.__rop_expr__("&", other)
+
+    def __xor__(self, other):
+        return self.__op_expr__("^", other)
+
+    def __rxor__(self, other):
+        return self.__rop_expr__("^", other)
+
+    def __or__(self, other):
+        return self.__op_expr__("|", other)
+
+    def __ror__(self, other):
+        return self.__rop_expr__("|", other)
+
+    # ~ is bitwise negation in Python
     # not/~ isn't applicable to objects
     # https://docs.python.org/3/library/operator.html
 
@@ -327,7 +328,7 @@ class Term(PreTerm, ABC):
         return self.__op_expr__("mod", other)
 
     def remainder(self, other):
-        return self.__op_expr__("remainder", other)
+        return self.__op_expr__("remainder", other, inline=False, method=True)
 
     def sqrt(self):
         return self.__uop_expr__("sqrt")
@@ -336,10 +337,10 @@ class Term(PreTerm, ABC):
         return self.__uop_expr__("abs")
 
     def maximum(self, other):
-        return self.__op_expr__("maximum", other)
+        return self.__op_expr__("maximum", other, method=True, inline=False)
 
     def minimum(self, other):
-        return self.__op_expr__("minimum", other)
+        return self.__op_expr__("minimum", other, method=True, inline=False)
 
     def fmax(self, other):
         return self.__op_expr__("fmax", other, inline=False)
@@ -453,13 +454,16 @@ class Term(PreTerm, ABC):
     def if_else(self, x, y):
         return self.__triop_expr__("if_else", x, y, method=True)
 
+    def is_in(self, x):
+        return self.__op_expr__("is_in", x, inline=False, method=True)
+
     def co_equalizer(self, x):
         return self.__op_expr__("co_equalizer", x, inline=False, method=True)
 
 
 class Value(Term):
     def __init__(self, value):
-        allowed = [int, float, str, bool]
+        allowed = [int, float, str, bool, ListTerm, tuple]
         if not any([isinstance(value, tp) for tp in allowed]):
             raise TypeError("value type must be one of: " + str(allowed))
         self.value = value
@@ -477,6 +481,10 @@ class Value(Term):
     def to_python(self, *, want_inline_parens=False):
         return self.value.__repr__()
 
+    # don't collect -5 as a complex expression
+    def __neg__(self):
+        return Value(-self.value)
+
 
 class UnQuotedStr(str):
     def __init__(self, v):
@@ -519,58 +527,11 @@ class FnValue(PreTerm):
         return UnQuotedStr(self.name)
 
 
-class FnCall(PreTerm):
-    # represent a function of columns
-    def __init__(self, value, fn_args=None, *, name=None):
-        if not callable(value):
-            raise TypeError("value type must be callable")
-        self.value = value
-        if name is None:
-            name = value.__name__
-        self.name = name
-        if fn_args is None:
-            fn_args = []
-        if isinstance(fn_args, ColumnReference):
-            fn_args = [fn_args]
-        for v in fn_args:
-            if not isinstance(v, ColumnReference):
-                raise TypeError("Expected fn_args to be None or all ColumnReference")
-        self.args = fn_args
-        self.display_form = (
-            name + "(" + ", ".join([fi.column_name for fi in fn_args]) + ")"
-        )
-        PreTerm.__init__(self)
-
-    def is_equal(self, other):
-        # can't use == as that builds a larger expression
-        if not isinstance(other, FnCall):
-            return False
-        if self.display_form != other.display_form:
-            return False
-        if len(self.args) != len(other.args):
-            return False
-        for li, ri in zip(self.args, other.args):
-            if not li.is_equal(ri):
-                return False
-        return True
-
-    def get_column_names(self, columns_seen):
-        for vi in self.args:
-            columns_seen.add(str(vi))
-
-    def replace_view(self, view):
-        self.args = [ai.replace_view(view) for ai in self.args]
-        return self
-
-    def to_python(self, *, want_inline_parens=False):
-        return UnQuotedStr(self.display_form)
-
-
 class ListTerm(PreTerm):
     def __init__(self, value):
         if not isinstance(value, list):
             raise TypeError("value type must be a list")
-        self.value = value
+        self.value = [v for v in value]  # standardize to a list
         PreTerm.__init__(self)
 
     def is_equal(self, other):
@@ -584,11 +545,17 @@ class ListTerm(PreTerm):
         return ListTerm(new_list)
 
     def to_python(self, *, want_inline_parens=False):
+        def li_to_python(value, *, want_inline_parens=False):
+            try:
+                return value.to_python(want_inline_parens=want_inline_parens)
+            except AttributeError:
+                return str(value)  # TODO: check if this should be repr?
+
         return (
             "["
             + ", ".join(
                 [
-                    ai.to_python(want_inline_parens=want_inline_parens)
+                    li_to_python(ai, want_inline_parens=want_inline_parens)
                     for ai in self.value
                 ]
             )
@@ -615,6 +582,8 @@ class ListTerm(PreTerm):
 def _enc_value(value):
     if isinstance(value, Term):
         return value
+    if isinstance(value, FnValue):
+        return value
     if callable(value):
         return FnValue(value)
     if isinstance(value, list):
@@ -635,7 +604,7 @@ class ColumnReference(Term):
             raise TypeError("column_name must be a string")
         if view is not None:
             if column_name not in view.column_set:
-                raise KeyError("column_name must be a column of the given view")
+                raise KeyError("column_name '" + str(column_name) + "' must be a column of the given view")
         Term.__init__(self)
 
     def is_equal(self, other):
@@ -744,45 +713,18 @@ class Expression(Term):
         return self.op + "(" + ", ".join(subs) + ")"
 
 
-# Some notes on trying to harden eval:
-#  http://lybniz2.sourceforge.net/safeeval.html
-
-
-def _parse_by_eval(source_str, *, data_def, outter_environemnt=None):
-    if not isinstance(source_str, str):
-        source_str = str(source_str)
-    if outter_environemnt is None:
-        outter_environemnt = {}
-    else:
-        outter_environemnt = {
-            k: v for (k, v) in outter_environemnt.items() if not k.startswith("_")
-        }
-    # don't have to completely kill this environment, as the code is something
-    # the user intends to run (and may have even typed in).
-    # But let's cut down the builtins anyway.
-    outter_environemnt["__builtins__"] = {
-        k: v for (k, v) in outter_environemnt.items() if isinstance(v, Exception)
-    }
-    v = eval(
-        source_str, outter_environemnt, data_def
-    )  # eval is eval(source, globals, locals)- so mp is first
-    if not isinstance(v, PreTerm):
-        v = _enc_value(v)
-    v.source_string = source_str
-    return v
-
-
 # define with def so function has usable __name__
 def connected_components(f, g):
     return data_algebra.expr_rep.Expression(op="connected_components", args=[f, g])
 
 
+# TODO: get rid of user_values
 def populate_specials(*, column_defs, destination, user_values=None):
     """populate a dictionary with special values
        column_defs is a dictionary,
-         usually formed from a ViewRepresentation.column_map.__dict__
+         usually formed from a ViewRepresentation.column_map
        destination is a dictionary,
-         usually formed from a ViewRepresentation.column_map.__dict__.copy()
+         usually formed from a ViewRepresentation.column_map.copy()
     """
 
     if not isinstance(column_defs, dict):
@@ -794,12 +736,9 @@ def populate_specials(*, column_defs, destination, user_values=None):
     if not isinstance(user_values, dict):
         raise TypeError("user_values should be a dictionary")
     nd = column_defs.copy()
-    ns = data_algebra.env.SimpleNamespaceDict(**nd)
-    # TODO: unify with custom_functions
+    # TODO: unify with custom_functions, and more doc!
     # makes these symbols available for parsing step
     # need to enter user functions here or as methods on variables
-    destination["_"] = ns
-    destination["_get"] = lambda key: user_values[key]
     # Note: a lot of the re-emitters add back the underbar, allow the non
     # underbar names we see here.  See expr_rep.py Expression to_python and to_pandas.
     destination["_row_number"] = lambda: data_algebra.expr_rep.Expression(
@@ -817,76 +756,13 @@ def populate_specials(*, column_defs, destination, user_values=None):
     )
 
 
-def parse_assignments_in_context(ops, view, *, parse_env=None):
-    """
-    Convert all entries of ops map to Term-expressions
-
-    Note: eval() is called to interpret expressions on some nodes, so this
-       function is not safe to use on untrusted code (though a somewhat restricted
-       version of eval() is used to try and catch some issues).
-    :param ops: dictionary from strings to expressions (either Terms or strings)
-    :param view: a data_algebra.data_ops.ViewRepresentation
-    :param parse_env map of names to values to add to parsing environment
-    :return:
-    """
-    if ops is None:
-        ops = {}
-    if isinstance(ops, tuple) or isinstance(ops, list):
-        opslen = len(ops)
-        ops = {k: v for (k, v) in ops}
-        if opslen != len(ops):
-            raise ValueError("ops keys must be unique")
-    if not isinstance(ops, dict):
-        raise TypeError("ops should be a dictionary")
-    column_defs = view.column_map.__dict__
-    if not isinstance(column_defs, dict):
-        raise TypeError("column_defs should be a dictionary")
-    if parse_env is None:
-        parse_env = data_algebra.env.outer_namespace()
-    if parse_env is None:
-        parse_env = {}
-    # first: make sure all entries are parsed
-    columns_used = set()
-    newops = collections.OrderedDict()
-    mp = column_defs.copy()
-    populate_specials(column_defs=column_defs, destination=mp, user_values=parse_env)
-    for k in ops.keys():
-        if not isinstance(k, str):
-            raise TypeError("ops keys should be strings")
-        ov = ops[k]
-        v = ov
-        if not isinstance(v, PreTerm):
-            if callable(v):
-                v = FnValue(v)
-            else:
-                v = _parse_by_eval(
-                    source_str=v, data_def=mp, outter_environemnt=parse_env
-                )
-        else:
-            v = v.replace_view(view)
-        newops[k] = v
-        used_here = set()
-        v.get_column_names(used_here)
-        columns_used = columns_used.union(
-            used_here - {k}
-        )  # can use a column to update itself
-    intersect = set(ops.keys()).intersection(columns_used)
-    if len(intersect) > 0:
-        raise ValueError(
-            "columns both produced and used in same expression set: " + str(intersect)
-        )
-    return newops
-
-
 def standardize_join_type(join_str):
     if not isinstance(join_str, str):
         raise TypeError("Expected join_str to be a string")
     join_str = join_str.upper()
-    re_map = {"OUTER": "FULL"}
-    try:
-        return re_map[join_str]
-    except KeyError:
-        pass
+    allowed = set(['INNER', 'LEFT', 'RIGHT', 'OUTER', 'FULL', 'CROSS'])
+    if not join_str in allowed:  # TOOO put this back in!
+        raise KeyError(f"join type {join_str} not supported")
     return join_str
 
 
