@@ -4,6 +4,7 @@ import numbers
 
 import numpy
 
+import data_algebra.user_fn
 import data_algebra.util
 import data_algebra.data_model
 import data_algebra.expr_rep
@@ -72,10 +73,6 @@ class PandasModelBase(data_algebra.data_model.DataModel, ABC):
             raise TypeError(
                 "op was supposed to be a data_algebra.data_ops.TableDescription"
             )
-        if len(op.qualifiers) > 0:
-            raise ValueError(
-                "table descriptions used with table_step() must not have qualifiers"
-            )
         df = data_map[op.table_name]
         if not self.is_appropriate_data_instance(df):
             raise ValueError("data_map[" + op.table_name + "] was not the right type")
@@ -107,10 +104,10 @@ class PandasModelBase(data_algebra.data_model.DataModel, ABC):
         standin_name = "_data_algebra_temp_g"  # name of an arbitrary input variable
         if not window_situation:
             for (k, opk) in op.ops.items():
-                if isinstance(opk, data_algebra.expr_rep.FnCall):
+                if isinstance(opk, data_algebra.user_fn.FnTerm):
                     # res[k] = opk.value(*[res[nm.column_name] for nm in opk.args])
                     pe = self.pandas_eval_env.copy()
-                    pe[opk.name] = opk.value
+                    pe[opk.name] = opk.pandas_fn
                     op_src = (
                         "@"
                         + opk.name
@@ -227,8 +224,8 @@ class PandasModelBase(data_algebra.data_model.DataModel, ABC):
         if len(op.ops) > 0:
             cols = {}
             for k, opk in op.ops.items():
-                if isinstance(opk, data_algebra.expr_rep.FnCall):
-                    fn = opk.value
+                if isinstance(opk, data_algebra.user_fn.FnTerm):
+                    fn = opk.pandas_fn
                     vk = res[str(opk.args[0])].agg(fn)
                 else:
                     if len(opk.args) > 0:
@@ -323,7 +320,10 @@ class PandasModelBase(data_algebra.data_model.DataModel, ABC):
         if not isinstance(jointype, str):
             raise TypeError("expected jointype to be a string")
         jointype = jointype.lower()
-        mp = {"full": "outer"}
+        mp = {
+            "full": "outer",
+            "cross": "outer"  # cross new to Pandas 1.2.0 December 2020
+        }
         try:
             return mp[jointype]
         except KeyError:
@@ -344,16 +344,27 @@ class PandasModelBase(data_algebra.data_model.DataModel, ABC):
         common_cols = set([c for c in left.columns]).intersection(
             [c for c in right.columns]
         )
+        by = op.by
+        if by is None:
+            by = []
+        scratch_col = None  # extra column to prevent empty-by issues
+        if len(by) <= 0:
+            scratch_col = 'data_algebra_temp_merge_col'
+            by = [scratch_col]
+            left[scratch_col] = 1
+            right[scratch_col] = 1
         # noinspection PyUnresolvedReferences
         res = self.pd.merge(
             left=left,
             right=right,
             how=self.standardize_join_code(op.jointype),
-            on=op.by,
+            on=by,
             sort=False,
             suffixes=("", "_tmp_right_col"),
         )
         res = res.reset_index(drop=True)
+        if scratch_col is not None:
+            del res[scratch_col]
         for c in common_cols:
             if c not in op.by:
                 is_null = res[c].isnull()
